
#ifndef PLS_CONT_MANAGER_H_
#define PLS_CONT_MANAGER_H_

#include <memory>
#include <utility>
#include <array>

#include "pls/internal/data_structures/aligned_stack.h"
#include "pls/internal/scheduling/continuation.h"

namespace pls {
namespace internal {
namespace scheduling {

class cont_manager {
 public:
  // Helper to pass the compile time constants to the constructor.
  template<size_t NUM_CONTS, size_t MAX_CONT_SIZE>
  struct template_args {};

  template<size_t NUM_CONTS, size_t MAX_CONT_SIZE>
  explicit cont_manager(data_structures::aligned_stack &cont_storage, template_args<NUM_CONTS, MAX_CONT_SIZE>)
      : max_cont_size_{MAX_CONT_SIZE} {

    // First node is currently active and our local start
    start_node_ = active_node_ = init_cont_node<MAX_CONT_SIZE>(cont_storage, nullptr, nullptr);

    // Build up chain after it
    continuation_node *current_node = start_node_;
    for (size_t i = 1; i < NUM_CONTS; i++) {
      continuation_node *next_node = init_cont_node<MAX_CONT_SIZE>(cont_storage, start_node_, current_node);
      current_node->next_ = next_node;
      current_node = next_node;
    }
  };

  // Fast-path methods
  continuation_node *fast_path_get_next() {
    auto result = active_node_;

    active_node_ = active_node_->next_;
    active_node_->cont_chain_start_ = start_node_;

    return result;
  }
  void fast_path_return() {
    active_node_ = active_node_->prev_;
  }

  // Slow-path methods
  void slow_path_return() {
    active_node_ = active_node_->prev_;
    active_node_->destroy_continuation();
  }
  bool falling_through() const {
    return fall_through_;
  }
  void fall_through() {
    fall_through_ = true;
    fall_through_node_ = nullptr;
  }
  void fall_through_and_execute(continuation_node *continuation_node) {
    fall_through_ = true;
    fall_through_node_ = continuation_node;
  }
  void aquire_clean_cont_chain(continuation_node *clean_chain) {
    // Link active node backwards and other chain forwards
    active_node_->prev_ = clean_chain;
    active_node_->cont_chain_start_ = clean_chain;
    clean_chain->next_ = active_node_;
    // Update our local chain start AND reset the active node to the start (we operate on a clean chain)
    start_node_ = clean_chain->cont_chain_start_;
    active_node_ = clean_chain;
  }
  void aquire_blocked_cont_chain(continuation_node *blocked_chain) {
    // Link active node backwards and other chain forwards
    active_node_->prev_ = blocked_chain;
    active_node_->cont_chain_start_ = blocked_chain;
    blocked_chain->next_ = active_node_;
    // Update our local chain start, NOT our active node (we continue execution on top of the blocking chain)
    start_node_ = blocked_chain->cont_chain_start_;
  }

  void execute_fall_through_continuation() {
    PLS_ASSERT(fall_through_node_ != nullptr,
               "Must not execute continuation node without associated continuation code.");

    // Reset our state to from falling through
    continuation_node *executed_node = fall_through_node_;
    fall_through_ = false;
    fall_through_node_ = nullptr;

    // Grab everyone involved in the transaction
    base_continuation *executed_continuation = executed_node->continuation_;
    base_continuation *parent_continuation = executed_continuation->parent_;
    continuation_node *parent_node = parent_continuation->node_;

    // Execute the continuation itself
    executed_continuation->execute();
    // Slow path return destroys it and resets our current cont_node
    slow_path_return();

    if (falling_through()) {
      return;
    }

    if (executed_node->is_end_of_cont_chain()) {
      // We are at the end of our local part of the cont chain and therefore
      // not invested in the continuation_chain above.

      // Notify the next continuation of finishing a child...
      if (parent_node->results_missing_.fetch_add(-1) == 1) {
        // ... we finished the last continuation, thus we need to take responsibility of the above cont chain.
        aquire_blocked_cont_chain(parent_node->cont_chain_start_);
        fall_through_and_execute(parent_node);
        return;
      } else {
        // ... we did not finish the last continuation, thus we are not concerned with this computational
        // path anymore. Just ignore it and fall through to mind our own business.
        fall_through();
        return;
      }
    } else {
      // We are invested in the continuation_chain above (its part of our local continuation node chain).
      // We keep executing it if possible and need to 'clean our state' with a swapped chain if not.

      // ...we could be blocked (and then need a clean cont chain).
      // For the rare case that the stealing processes is still going on, we
      // need to fight with it over directly executing the task that is currently being stolen.
      continuation_node::stamped_state task_state{continuation_node::state::execute_local};
      parent_node->state_.exchange(task_state, std::memory_order_acq_rel);
      if (task_state.value == continuation_node::state::stolen) {
        // The other task strain is executing.
        // Notify the parent continuation of finishing our child...
        if (parent_node->results_missing_.fetch_add(-1) == 1) {
          // ... we finished the last continuation, thus we keep the responsibility for the above cont chain.
          fall_through_and_execute(parent_node);
          return;
        } else {
          // ... we did not finish the last continuation, thus we need to give up our local cont chain in
          // favour of the stolen one (leaving us with a clean chain for further operation).
          aquire_clean_cont_chain(parent_node->cont_chain_start_);
          fall_through();
          return;
        }
      } else {
        // We 'overruled' the stealing thread and execute the other task ourselfs.
        // We can be sure that only we where involved in executing the child tasks of the parent_continuation...
        parent_continuation->node_->results_missing_.fetch_add(-1);
        parent_continuation->execute_task();
        if (falling_through()) {
          // ... if the second strain was interrupted we fall through without scheduling the parent_continuation
          // (the currently pending/interrupted strain will do this itself).
          return;
        }
        // ...if we could execute the second branch without interruption,
        // we can schedule the parent safely for execution.
        fall_through_and_execute(parent_continuation->node_);
      }
    }
  }

  size_t get_max_cont_size() const { return max_cont_size_; }

 private:
  template<size_t MAX_CONT_SIZE>
  static continuation_node *init_cont_node(data_structures::aligned_stack &cont_storage,
                                           continuation_node *cont_chain_start,
                                           continuation_node *prev) {
    // Represents one cont_node and its corresponding memory buffer (as one continuous block of memory).
    constexpr size_t buffer_size = MAX_CONT_SIZE - sizeof(continuation_node);
    char *cont_node_address = cont_storage.push_bytes<continuation_node>();
    char *cont_node_memory_address = cont_storage.push_bytes(buffer_size);

    return new(cont_node_address) continuation_node(cont_node_memory_address, cont_chain_start, prev);
  }

 private:
  const size_t max_cont_size_;

  /**
   * Managing the continuation chain.
   */
  continuation_node *start_node_;
  continuation_node *active_node_;

  /**
   * Managing falling through back to the scheduler.
   */
  bool fall_through_{false};
  continuation_node *fall_through_node_{nullptr};
};

template<size_t NUM_CONTS, size_t MAX_CONT_SIZE>
class static_cont_manager {
 public:
  static_cont_manager()
      : static_cont_storage_{},
        cont_manager_(static_cont_storage_.get_stack(), cont_manager::template_args<NUM_CONTS, MAX_CONT_SIZE>{}) {}
  cont_manager &get_cont_manager() { return cont_manager_; }

 private:
  data_structures::static_aligned_stack<NUM_CONTS * MAX_CONT_SIZE> static_cont_storage_;
  cont_manager cont_manager_;
};

}
}
}
#endif //PLS_CONT_MANAGER_H_
