
#ifndef PLS_INTERNAL_SCHEDULING_CONTINUATION_H_
#define PLS_INTERNAL_SCHEDULING_CONTINUATION_H_

#include <type_traits>
#include <atomic>
#include <utility>

#include "pls/internal/data_structures/stamped_integer.h"
#include "pls/internal/data_structures/delayed_initialization.h"
#include "pls/internal/base/alignment.h"
#include "parallel_result.h"

namespace pls {
namespace internal {
namespace scheduling {

class continuation_node;

class base_continuation {
  friend class cont_manager;
 protected:
  // We plan to only init the members for a continuation on the slow path.
  // If we can execute everything inline we simply skip it saving runtime overhead.
  template<typename T>
  using delayed_init = data_structures::delayed_initialization<T>;

 public:
  explicit base_continuation(base_continuation *parent, continuation_node *node, unsigned int result_index = 0)
      : parent_{parent},
        node_{node},
        result_index_{result_index} {}

  virtual void execute() = 0;
  virtual void execute_task() = 0;
  virtual ~base_continuation() = default;

  virtual void *get_result_pointer(unsigned short index) = 0;
  template<typename T>
  void store_result(unsigned short index, T &&result) {
    using BASE_T = typename std::remove_cv<typename std::remove_reference<T>::type>::type;
    reinterpret_cast<delayed_init<BASE_T> *>(get_result_pointer(index))->initialize(std::forward<T>(result));
  }

  base_continuation *get_parent() { return parent_; }
  continuation_node *get_cont_node() { return node_; }

 protected:
  base_continuation *parent_;
  continuation_node *node_;
  unsigned int result_index_;
};

class continuation_node {
  friend class cont_manager;

 public:
  continuation_node(char *memory, continuation_node *cont_chain_start, continuation_node *prev)
      : cont_chain_start_{cont_chain_start},
        prev_{prev},
        memory_{memory} {}

  // Management of the associated continuation
  template<typename T, typename ...ARGS>
  T *init_continuation(ARGS &&...args) {
    PLS_ASSERT(continuation_ == nullptr, "Must only allocate one continuation at once per node.")

    auto *result = new(memory_) T(std::forward<ARGS>(args)...);
    continuation_ = result;
    return result;
  }
  void destroy_continuation() {
    // Deconstruct Continuation
    continuation_->~base_continuation();
    continuation_ = nullptr;

    // Reset Associated counters
    results_missing_.store(2);
    offered_chain_.store(nullptr);
    auto old_state = state_.load();
    state_.store({old_state.stamp + 1, initialized});
  }
  template<typename T>
  void destroy_continuation_fast() {
    (*reinterpret_cast<T *>(continuation_)).~T();
    continuation_ = nullptr;
  }
  base_continuation *get_continuation() {
    return continuation_;
  }
  continuation_node *get_prev() {
    return prev_;
  }

  bool is_end_of_cont_chain() {
    return prev_ == continuation_->get_parent()->get_cont_node();
  }

 private:
  // Linked list property of continuations (continuation chains as memory management).
  // Each continuation knows its chain start to allow stealing a whole chain in O(1)
  // without the need to traverse back to the chain start.
  continuation_node *cont_chain_start_;
  continuation_node *prev_, *next_{nullptr};

  // When blocked on this continuation, we need to know what other chain we
  // got offered by the stealing thread.
  // For this we need only the head of the other chain (as each continuation is a
  // self describing entity for its chain up to the given node).
  std::atomic<continuation_node *> offered_chain_{nullptr};

  // Management for coordinating concurrent result writing and stealing.
  // The result count decides atomically who gets to execute the continuation.
  std::atomic<unsigned short> results_missing_{2};

  // The flag is needed for an ongoing stealing request.
  // Stealing threads need to offer their continuation chain before the
  // 'fully' own the stolen task. As long as that is not done the continuation
  // chain can abort the steal request in order to be not blocked without a
  // new, clean continuation chain to work with.
  enum state { initialized, execute_local, stealing, stolen };
  using stamped_state = data_structures::stamped_integer;
  std::atomic<stamped_state> state_{{initialized}};

  // Pointer to memory region reserved for the companion continuation.
  // Must be a buffer big enough to hold any continuation encountered in the program.
  // This memory is managed explicitly by the continuation manager and runtime system
  // (they need to make sure to always call de-constructors and never allocate two continuations).
  char *memory_;
  base_continuation *continuation_{nullptr};
};

template<typename T2, typename R1, typename R2, typename F>
class continuation : public base_continuation {
 private:
  template<typename RES_TYPE>
  struct result_runner {
    // Strip off unwanted modifiers...
    using BASE_RES_TYPE = typename std::remove_cv<typename std::remove_reference<RES_TYPE>::type>::type;

    static void execute(continuation &cont) {
      parallel_result<BASE_RES_TYPE> result{cont.function_((*cont.result_1_).value(), (*cont.result_2_).value())};
      if (result.fast_path()) {
        cont.parent_->store_result(cont.result_index_, std::move(result));
      }
    }
  };
  template<typename INNER_TYPE>
  struct result_runner<parallel_result<INNER_TYPE>> {
    static void execute(continuation &cont) {
      auto result = cont.function_((*cont.result_1_).value(), (*cont.result_2_).value());
      if (result.fast_path()) {
        cont.parent_->store_result(cont.result_index_, std::move(result));
      }
    }
  };

 public:
  template<typename FARG, typename ...T2ARGS>
  explicit continuation(base_continuation *parent,
                        continuation_node *node,
                        unsigned int result_index,
                        FARG &&function,
                        T2ARGS...task_2_args):
      base_continuation(parent, node, result_index),
      function_{std::forward<FARG>(function)},
      task_{std::forward<T2ARGS>(task_2_args)...} {}
  ~continuation() override = default;

  void execute() override {
    using result_type = decltype(function_((*result_1_).value(), (*result_2_).value()));
    result_runner<result_type>::execute(*this);
  }

  void execute_task() override {
    task_.execute();
  }

  void *get_result_pointer(unsigned short index) override {
    switch (index) {
      case 0:return &result_1_;
      case 1:return &result_2_;
      default: PLS_ERROR("Unexpected Result Index!")
    }
  }

  T2 *get_task() {
    return &task_;
  }

 private:
  // Initial data members. These slow down the fast path, try to init them lazy when possible.
  F function_;
  T2 task_;

  // Some fields/actual values stay uninitialized (save time on the fast path if we don not need them).
  // More fields untouched on the fast path is good, but for ease of an implementation we only keep some for now.
  delayed_init<R1> result_1_;
  delayed_init<R2> result_2_;
};

}
}
}

#endif //PLS_INTERNAL_SCHEDULING_CONTINUATION_H_
