
#ifndef PLS_SCHEDULER_IMPL_H
#define PLS_SCHEDULER_IMPL_H

#include <utility>
#include "pls/internal/scheduling/continuation.h"
#include "pls/internal/scheduling/parallel_result.h"
#include "pls/internal/scheduling/task.h"

namespace pls {
namespace internal {
namespace scheduling {

template<typename F1, typename F2>
struct scheduler::starter {
  F1 function_1_;
  F2 function_2_;
  using return_type_1 = decltype(function_1_());
  using return_type_2 = decltype(function_2_());

  template<typename F1ARG, typename F2ARG>
  explicit starter(F1ARG &&function_1, F2ARG &&function_2) : function_1_{std::forward<F1ARG>(function_1)},
                                                             function_2_{std::forward<F2ARG>(function_2)} {};

  template<typename FCONT>
  auto then(FCONT &&cont_function)
  -> decltype(cont_function(std::declval<typename return_type_1::value_type>(),
                            std::declval<typename return_type_2::value_type>())) {
    using continuation_type = continuation<task<F2>, return_type_1, return_type_2, FCONT>;
    using result_type = decltype(cont_function(std::declval<typename return_type_1::value_type>(),
                                               std::declval<typename return_type_2::value_type>()));

    auto &my_state = thread_state::get();
    auto &cont_manager = my_state.get_cont_manager();

    PLS_ASSERT(sizeof(continuation_type) <= cont_manager.get_max_cont_size(),
               "Must stay within the size limit of the static memory configuration.");

    // Select current continuation.
    // For now directly copy both the continuation function and the second task.
    auto *current_cont_node = cont_manager.fast_path_get_next();

    // TODO: Fix null pointers at very first spawn...
    // In the fast path case we are always on the left side of the tree
    // and our prev cont chain node always also holds our parent continuation.
    base_continuation *parent_cont =
        current_cont_node->get_prev() == nullptr ? nullptr : current_cont_node->get_prev()->get_continuation();
    unsigned short int result_index = 0;
    auto *current_cont = current_cont_node->init_continuation<continuation_type>(parent_cont,
                                                                                 current_cont_node,
                                                                                 result_index,
                                                                                 cont_function,
                                                                                 function_2_,
                                                                                 current_cont_node);

    // Publish the second task.
    my_state.get_task_manager().publish_task(current_cont->get_task());

    // Call first function on fast path
    return_type_1 result_1 = function_1_();
    if (cont_manager.falling_through()) {
      return result_type{}; // Unwind stack...
    }

    // Try to call second function on fast path
    if (my_state.get_task_manager().steal_local_task()) {
      return_type_2 result_2 = function_2_();
      if (cont_manager.falling_through()) {
        return result_type{}; // Unwind stack...
      }

      // We fully got all results, inline as good as possible.
      // This is the common case, branch prediction should be rather good here.

      // Just return the cont object unused and directly call the function.
      current_cont_node->destroy_continuation_fast<continuation_type>();
      my_state.get_cont_manager().fast_path_return();

      auto cont_result = cont_function(result_1.value(), result_2.value());
      if (cont_manager.falling_through()) {
        return result_type{}; // Unwind stack...
      }
      return cont_result;
    }

    cont_manager.fall_through();
    return result_type{};
  };
};

template<typename F1, typename F2>
scheduler::starter<F1, F2> scheduler::invoke_parallel(F1 &&function_1, F2 &&function_2) {
  return scheduler::starter<F1, F2>{std::forward<F1>(function_1), std::forward<F2>(function_2)};
}

template<typename F1, typename F2>
scheduler::starter<F1, F2> scheduler::par(F1 &&function_1, F2 &&function_2) {
  return thread_state::get().get_scheduler().invoke_parallel(std::forward<F1>(function_1),
                                                             std::forward<F2>(function_2));
}

class scheduler::init_function {
 public:
  virtual parallel_result<int> run() = 0;
};
template<typename F>
class scheduler::init_function_impl : public init_function {
 public:
  explicit init_function_impl(F &function) : function_{function} {}
  parallel_result<int> run() override {
    return scheduler::par([]() {
      // No-op
      return parallel_result<int>{0};
    }, [=]() {
      function_();
      return parallel_result<int>{0};
    }).then([](const int &, const int &) {
      // Notify that work is done after finishing the last user continuation.
      thread_state::get().get_scheduler().work_section_done_ = true;
      return parallel_result<int>{0};
    });
  }
 private:
  F &function_;
};

template<typename Function>
void scheduler::perform_work(Function work_section) {
  PROFILE_WORK_BLOCK("scheduler::perform_work")

  // Prepare main root task
  init_function_impl<Function> starter_function{work_section};
  main_thread_starter_function_ = &starter_function;

  work_section_done_ = false;
  if (reuse_thread_) {
    auto &my_state = memory_.thread_state_for(0);
    base::this_thread::set_state(&my_state); // Make THIS THREAD become the main worker

    sync_barrier_.wait(); // Trigger threads to wake up
    work_thread_work_section(); // Simply also perform the work section on the main loop
    sync_barrier_.wait(); // Wait for threads to finish
  } else {
    // Simply trigger the others to do the work, this thread will sleep/wait for the time being
    sync_barrier_.wait(); // Trigger threads to wake up
    sync_barrier_.wait(); // Wait for threads to finish
  }
}

}
}
}

#endif //PLS_SCHEDULER_IMPL_H
